package com.trafficparrot.sdk.file;

import com.github.jknack.handlebars.Handlebars;
import com.github.jknack.handlebars.HandlebarsException;
import com.github.jknack.handlebars.Template;
import com.google.common.collect.ImmutableMap;
import com.trafficparrot.sdk.jms.TpLogger;

import java.io.IOException;
import java.util.Map;

import static com.trafficparrot.sdk.handlebars.HandlebarsFactory.handlebars;
import static java.lang.String.format;
import static java.util.Collections.singletonMap;

public class FileMessageHandlebarsResponseTransformer {
    private final Handlebars handlebars;
    private final boolean enabled;
    private final TpLogger logger;

    public FileMessageHandlebarsResponseTransformer(Handlebars handlebars, boolean enabled, TpLogger logger) {
        this.handlebars = handlebars;
        this.enabled = enabled;
        this.logger = logger;
    }

    public static FileMessageHandlebarsResponseTransformer handlebarsTransformer(String commaSeparatedHandlebarHelpers, Boolean handlebarsEnabled, TpLogger logger) {
        if (handlebarsEnabled) {
            return defineHelpers(logger, handlebars(commaSeparatedHandlebarHelpers));
        }
        return handlebarsAreDisabled(logger);
    }

    public static FileMessageHandlebarsResponseTransformer handlebarsAreDisabled(TpLogger logger) {
        return new FileMessageHandlebarsResponseTransformer(null, false, logger);
    }

    private static FileMessageHandlebarsResponseTransformer defineHelpers(TpLogger logger, Handlebars handlebars) {
        return new FileMessageHandlebarsResponseTransformer(handlebars, true, logger);
    }

    public FileMessageResponse transform(FileMessage request, FileMessageResponse response) {
        if (!enabled) {
            return response;
        }
        Map<String, ?> requestModel = requestModel(request);

        String responseTextWithTemplateApplied;
        try {
            responseTextWithTemplateApplied = applyTemplate(requestModel, response.text);
        } catch (HandlebarsException | IOException e) {
            return errorResponse(response.text, response, e);
        }

        String fileNameWithTemplateApplied;
        try {
            fileNameWithTemplateApplied = applyTemplate(requestModel, response.fileName);
        } catch (HandlebarsException | IOException e) {
            return errorResponse(response.fileName, response, e);
        }
        return new FileMessageResponse(response.destination, response.fileMessageResponseTransformerClassName, fileNameWithTemplateApplied, responseTextWithTemplateApplied, response.bodyType);
    }

    private FileMessageResponse errorResponse(String original, FileMessageResponse response, Exception e) {
        String errorMessage = format("Problem while applying handlebars: \n\n%s\n\nOriginal text: \n%s", e.getMessage(), original);
        logger.error(errorMessage, e);
        return new FileMessageResponse(response.destination, response.fileName, errorMessage + ". \n\nSee logs for more details.");
    }

    private String applyTemplate(Map<String, ?> requestModel, String text) throws IOException {
        Template responseTemplate = handlebars.compileInline(text);
        return responseTemplate.apply(requestModel);
    }

    private Map<String, ?> requestModel(FileMessage requestMessage) {
        return singletonMap("request", ImmutableMap.of(
                "body", requestMessage.text,
                "fileName", requestMessage.fileName));
    }
}
