package com.trafficparrot.sdk.messaging;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Optional;
import com.google.common.base.Predicate;
import com.google.common.collect.Sets;
import com.trafficparrot.sdk.jms.Slf4jTpLogger;
import com.trafficparrot.sdk.jms.TpLogger;
import net.bytebuddy.ByteBuddy;
import net.bytebuddy.description.modifier.ModifierContributor.ForField;
import net.bytebuddy.description.modifier.SyntheticState;
import net.bytebuddy.description.type.TypeDescription;
import net.bytebuddy.description.type.TypeDescription.ForLoadedType;
import net.bytebuddy.dynamic.DynamicType;
import net.bytebuddy.dynamic.TargetType;
import net.bytebuddy.dynamic.scaffold.InstrumentedType;
import org.unsynchronized.jdeserialize.*;

import java.io.*;
import java.lang.reflect.InvocationHandler;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.regex.Pattern;

import static com.google.common.base.Preconditions.checkState;
import static com.google.common.base.Predicates.or;
import static com.google.common.collect.Iterables.any;
import static com.trafficparrot.sdk.messaging.FindAndReplaceBytes.findAndReplaceShortUtf;
import static java.io.ObjectStreamConstants.SC_WRITE_METHOD;
import static java.lang.String.format;
import static java.nio.file.FileVisitResult.CONTINUE;
import static java.nio.file.Files.walkFileTree;
import static java.util.Arrays.asList;
import static java.util.Collections.emptySet;
import static java.util.Collections.singleton;
import static java.util.regex.Pattern.quote;
import static org.apache.commons.lang3.StringUtils.join;
import static org.unsynchronized.jdeserialize.fieldtype.ARRAY;
import static org.unsynchronized.jdeserialize.fieldtype.OBJECT;

public class ObjectSerializedRepresentation {
    private static final TpLogger LOGGER = new Slf4jTpLogger(ObjectSerializedRepresentation.class);
    private static final String GENERATED_BY_TRAFFIC_PARROT = "$TrafficParrot$";
    private static final Lock CLASS_LOADING_LOCK = new ReentrantLock();
    private static final File CLASS_FILE_LOCATION = new File("lib" + File.separator + "trafficparrot" + File.separator + "generated"); // TODO(3/5) unhardcode?

    public static ObjectSerializedRepresentation OBJECT_SERIALIZED_REPRESENTATION = new ObjectSerializedRepresentation(LOGGER, CLASS_FILE_LOCATION);

    private final ConcurrentHashMap<String, Class<?>> classAliases = new ConcurrentHashMap<>();

    private final Set<String> nullClasses = Sets.newConcurrentHashSet();
    private final Set<String> notNullClasses = Sets.newConcurrentHashSet();
    private final ConcurrentHashMap<String, String> enclosingClasses = new ConcurrentHashMap<>();
    private final Set<instance> seen = Sets.newConcurrentHashSet();

    private final TpLogger logger;
    private final File classFileLocation;

    @VisibleForTesting
    ObjectSerializedRepresentation(TpLogger logger, File classFileLocation) {
        this.logger = logger;
        this.classFileLocation = classFileLocation.getAbsoluteFile();
        logger.info("Storing dynamic class files in: " + classFileLocation.getAbsolutePath());
    }

    public void loadGeneratedClasses() throws IOException {
        Files.createDirectories(classFileLocation.toPath());
        walkFileTree(classFileLocation.toPath(), new SimpleFileVisitor<Path>() {
            @Override
            public FileVisitResult visitFile(Path file, BasicFileAttributes attributes) throws IOException {
                if (attributes.isRegularFile()) {
                    loadGeneratedAlias(file);
                }
                return CONTINUE;
            }
        });
    }

    private void loadGeneratedAlias(Path classFile) {
        String internalClassName = className(classFile);
        String externalClassName = externalClassName(internalClassName);
        if (!internalClassName.equals(externalClassName)) {
            try {
                classAliases.put(externalClassName, Class.forName(internalClassName));
                logger.info("Loaded: " + externalClassName);
            } catch (ClassNotFoundException e) {
                logger.error("Could not load: " + externalClassName, e);
            }
        }
    }

    @VisibleForTesting
    String className(Path classFile) {
        return classFile.toAbsolutePath()
                .toString()
                .replaceFirst(".*" + Pattern.quote(classFileLocation.getPath() + File.separator), "")
                .replace(".class", "")
                .replace(File.separatorChar, '.');
    }

    public String internalClassName(String value) {
        Class<?> alias = classAliases.get(value);
        if (alias != null) {
            return alias.getName();
        } else {
            return value;
        }
    }

    public String externalClassName(String name) {
        return name.replaceAll(quote(GENERATED_BY_TRAFFIC_PARROT) + ".*", "");
    }

    public byte[] writeObjectBytes(Object object) throws IOException {
        try (ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
             ObjectOutputStream objectOutputStream = new ObjectOutputStream(byteArrayOutputStream)) {
            objectOutputStream.writeObject(object);
            byte[] bytes = byteArrayOutputStream.toByteArray();
            return replaceGeneratedWithActual(bytes);
        }
    }

    public Optional<Object> tryToReadObject(byte[] bytes) {
        try {
            return Optional.of(bytesToJavaObject(bytes));
        } catch (InvalidClassException e) {
            logger.error("The serialized object class is invalid", e);
            return Optional.absent();
        } catch (ClassNotFoundException | InvalidObjectException original) {
            try {
                return Optional.of(tryToReadShadowObject(bytes));
            } catch (IOException | ClassNotFoundException shadowClassNotFoundException) {
                ClassNotFoundException withCause = new ClassNotFoundException(original.getMessage(), shadowClassNotFoundException);
                logger.error("The serialized object class is not on the Traffic Parrot classpath and could not be dynamically loaded", withCause);
                return Optional.absent();
            }
        } catch (EOFException | StreamCorruptedException e) {
            // message does not look like a serialized object so do not clutter the logs with these exceptions
            LOGGER.debug("Does not look like this is a serialized Java object", e);
            return Optional.absent();
        } catch (IOException e) {
            logger.error("Problem while reading object", e);
            return Optional.absent();
        }
    }

    private Object tryToReadShadowObject(byte[] bytes) throws IOException, ClassNotFoundException {
        jdeserialize jdeserialize = new jdeserialize();
        jdeserialize.run(new ByteArrayInputStream(bytes), true);

        content instance = jdeserialize.getContent().get(0);

        CLASS_LOADING_LOCK.lock(); // only load one shadow object at a time to avoid race conditions
        try {
            seen.clear();
            nullClasses.clear();
            notNullClasses.clear();
            enclosingClasses.clear();
            loadValue(instance);
            byte[] generated = replaceActualWithGenerated(bytes);
            return bytesToJavaObject(generated);
        } finally {
            CLASS_LOADING_LOCK.unlock();
        }
    }

    private byte[] replaceGeneratedWithActual(byte[] bytes) throws IOException {
        for (String className : classAliases.keySet()) {
            String alias = classAliases.get(className).getName();
            bytes = findAndReplaceClass(bytes, alias, className);
        }
        return bytes;
    }

    // TODO(3/5) if the original object had a string with the class name, it will incorrectly be replaced by the generated name
    private byte[] replaceActualWithGenerated(byte[] bytes) throws IOException {
        for (String className : classAliases.keySet()) {
            String alias = classAliases.get(className).getName();
            bytes = findAndReplaceClass(bytes, className, alias);
        }
        return bytes;
    }

    private byte[] findAndReplaceClass(byte[] bytes, String className, String alias) throws IOException {
        byte[] original = bytes;
        bytes = findAndReplaceShortUtf(bytes, arrayFieldDescriptor(className), arrayFieldDescriptor(alias));
        bytes = findAndReplaceShortUtf(bytes, arrayClassName(className), arrayClassName(alias));
        if (!Arrays.equals(original, bytes)) {
            // cannot be both array and object so we can exit now to prevent overlap with the other replaces
            return bytes;
        }
        bytes = findAndReplaceShortUtf(bytes, fieldDescriptor(className), fieldDescriptor(alias));
        bytes = findAndReplaceShortUtf(bytes, className, alias);
        return bytes;
    }

    private void loadInstance(instance instance) throws ClassNotFoundException, IOException {
        for (classdesc innerClass : instance.classdesc.innerclasses) {
            enclosingClasses.put(innerClass.name, instance.classdesc.name);
        }
        switch (instance.classdesc.classtype) {
            case NORMALCLASS:
                loadNormalInstance(instance, false);
                break;
            case PROXYCLASS:
                loadProxyInstance(instance);
                break;
            default:
                throw new ClassNotFoundException("Unknown class type: " + instance.classdesc.classtype);
        }
    }

    private void loadProxyInstance(instance instance) throws ClassNotFoundException, IOException {
        Collection<Map<field, Object>> instanceFieldData = instance.fielddata.values();
        checkState(instanceFieldData.size() == 1);

        Collection<Object> fieldValues = instanceFieldData.iterator().next().values();
        checkState(fieldValues.size() == 1);

        instance handler = (instance) fieldValues.iterator().next();

        if (!instance.classdesc.annotations.isEmpty()) {
            throw new ClassNotFoundException(format("Proxy class handler '%s' could not be loaded since it has been annotated with a custom annotateProxyClass method", handler.classdesc.name));
        }

        for (String proxyInterface : proxyInterfaces(instance.classdesc)) {
            loadInterface(proxyInterface);
        }
        loadNormalInstance(handler, true);
    }

    private void loadNormalInstance(instance instance, boolean isInvocationHandler) throws ClassNotFoundException, IOException {
        if (seen.contains(instance)) {
            throw new ClassNotFoundException(format("Class '%s' could not be loaded since it is part of a circular dependency", instance.classdesc.name));
        }
        seen.add(instance);

        for (Map<field, Object> classFields : instance.fielddata.values()) {
            for (Map.Entry<field, Object> entry : classFields.entrySet()) {
                field field = entry.getKey();
                Object value = entry.getValue();
                if (field.type == OBJECT) {
                    String fieldClassName = className(field.classname.value);
                    if (value == null) {
                        nullClasses.add(fieldClassName);
                    } else {
                        notNullClasses.add(fieldClassName);
                    }
                } else if (field.type == ARRAY) {
                    arrayobj array = (arrayobj) value;
                    String fieldComponentClassName = className(stripArrayPrefix(field.classname.value));
                    if (alwaysNull(array)) {
                        nullClasses.add(fieldComponentClassName);
                    } else {
                        notNullClasses.add(fieldComponentClassName);
                    }
                }

                if (innerClassReferenceAlreadyLoaded(field, value)) {
                    continue;
                }

                loadValue(value);
            }
        }
        notNullClasses.add(instance.classdesc.name); // we are about to generate this one
        nullClasses.removeAll(notNullClasses);
        for (String nullClass : nullClasses) {
            loadNullClass(nullClass);
        }
        loadClass(instance.classdesc, isInvocationHandler);
    }

    private boolean innerClassReferenceAlreadyLoaded(field field, Object value) {
        return value instanceof instance
                && (field.isInnerClassReference() || enclosingClasses.containsKey(className(field.classname.value)))
                && seen.contains(value);
    }

    private boolean alwaysNull(arrayobj array) {
        for (Object data : array.data) {
            if (data != null) {
                return false;
            }
        }
        return true;
    }

    private void loadValue(Object value) throws ClassNotFoundException, IOException {
        if (value instanceof enumobj) {
            enumobj enumobj = (enumobj) value;
            loadEnum(enumobj.classdesc);
        } else if (value instanceof instance) {
            instance fieldInstance = (instance) value;
            loadInstance(fieldInstance);
        } else if (value instanceof arrayobj) {
            arrayobj arrayobj = (arrayobj) value;
            for (Object element : arrayobj.data) {
                loadValue(element);
            }
        } else if (value instanceof content && !(value instanceof stringobj)) {
            throw new UnsupportedOperationException("Unsupported field type: " + value.getClass());
        }
    }

    private void loadInterface(String name) throws IOException {
        Optional<? extends Class<?>> alreadyLoadedClass = alreadyLoadedClass(name);
        if (alreadyLoadedClass.isPresent()) {
            // we are never going to make things better by reloading an already loaded interface
            return;
        }

        load(new ByteBuddy()
                .makeInterface()
                .name(name));
    }

    private void loadNullClass(String name) throws IOException {
        Optional<? extends Class<?>> alreadyLoadedClass = alreadyLoadedClass(name);
        if (alreadyLoadedClass.isPresent()) {
            // we are never going to make things better by reloading an already loaded class as a null class
            return;
        }

        if (classAliases.containsKey(name)) {
            // already have an always null class definition
            return;
        }

        String alwaysNullClassName = alwaysNullClassName(name);
        Class<?> loaded = load(new ByteBuddy()
                .subclass(Object.class)
                .implement(Serializable.class)
                .name(alwaysNullClassName));

        recordAlias(name, loaded);

        logger.warn(format("Class '%s' was partially loaded. Only null values are supported for fields of this type since Traffic Parrot did not see the serialized class definition.", name));
    }

    private String alwaysNullClassName(String name) {
        return name + GENERATED_BY_TRAFFIC_PARROT + "AlwaysNull";
    }

    private void loadEnum(classdesc classdesc) throws ClassNotFoundException, IOException {
        Optional<? extends Class<?>> alreadyLoadedClass = alreadyLoadedClass(classdesc.name);
        if (alreadyLoadedClass.isPresent()) {
            // we have seen this class before so we trust it is fully loaded already
            return;
        }

        Set<String> existingEnumConstants = existingEnumConstants(classdesc);
        Set<String> newConstants = Sets.difference(classdesc.enumconstants, existingEnumConstants);
        if (newConstants.isEmpty()) {
            // we already have enough constants loaded
            return;
        }

        Set<String> allConstants = Sets.union(existingEnumConstants, classdesc.enumconstants);
        String alias = generateClassAlias(classdesc.name);

        Class<?> loaded = load(new ByteBuddy()
                .makeEnumeration(allConstants)
                .name(alias));

        recordAlias(classdesc.name, loaded);

        logger.info(format("Enum '%s' was loaded successfully with constants %s", classdesc.name, allConstants));
    }

    private void deleteClassFile(String name) {
        Path toDelete = classFileLocation.toPath().resolve(name.replace('.', File.separatorChar) + ".class");
        try {
            Files.delete(toDelete);
        } catch (IOException e) {
            logger.warn("Could not delete class file: " + toDelete, e);
        }
    }

    private Class<?> load(DynamicType.Builder<?> byteBuddy) throws IOException {
        DynamicType.Unloaded<?> make = byteBuddy.make();
        make.saveIn(classFileLocation);
        return make.load(getClass().getClassLoader()).getLoaded();
    }

    private Set<String> existingEnumConstants(classdesc classdesc) throws ClassNotFoundException {
        Set<String> existingConstants = new HashSet<>();
        if (classAliases.containsKey(classdesc.name)) {
            Class<?> existing = classAliases.get(classdesc.name);
            Object[] enumConstants = existing.getEnumConstants();

            for (Object constant : enumConstants) {
                existingConstants.add(constant.toString());
            }
        }
        return existingConstants;
    }

    private Class<?> loadClass(classdesc classdesc, boolean isInvocationHandler) throws ClassNotFoundException, IOException {
        String originalClassName = classdesc.isInnerClass() ? appendOuterClassNames(classdesc.name) : classdesc.name;
        Optional<? extends Class<?>> alreadyLoadedClass = alreadyLoadedClass(originalClassName);
        if (alreadyLoadedClass.isPresent()) {
            // we have seen this class before so we trust it is fully loaded already
            return alreadyLoadedClass.get();
        }

        if (!classdesc.annotations.isEmpty()) {
            throw new ClassNotFoundException(format("Class '%s' could not be loaded since it has been annotated with a custom annotateClass method", originalClassName));
        }

        if ((classdesc.descflags & SC_WRITE_METHOD) == SC_WRITE_METHOD) {
            throw new ClassNotFoundException(format("Class '%s' could not be loaded since it has a custom writeObject method", originalClassName));
        }

        boolean hasPartiallyLoadedFields = hasPartiallyLoadedFields(classdesc);
        String className = hasPartiallyLoadedFields ? generateClassAlias(originalClassName) : originalClassName;

        Class<?> superClass = superClass(classdesc);
        DynamicType.Builder<?> typeBuilder = new ByteBuddy()
                .subclass(superClass)
                .implement(Serializable.class)
                .name(className)
                .serialVersionUid(classdesc.serialVersionUID);

        if (isInvocationHandler) {
            typeBuilder = typeBuilder.implement(InvocationHandler.class);
        }

        for (field classField : classFields(classdesc)) {
            if (classField.type == OBJECT) {
                String fieldClass = className(classField.classname.value);
                classField.classname.value = fieldDescriptor(appendOuterClassNames(fieldClass));
            }
            typeBuilder = typeBuilder.defineField(classField.name, typeOf(className, classField), fieldModifiers(classField));
        }

        Class<?> loaded = load(typeBuilder);
        if (hasPartiallyLoadedFields) {
            recordAlias(originalClassName, loaded);
            logger.warn(format("Class '%s' was partially loaded", originalClassName));
        } else {
            removeAlias(originalClassName);
            logger.info(format("Class '%s' was loaded successfully", originalClassName));
        }
        return loaded;
    }

    private Set<? extends ForField> fieldModifiers(field classField) {
        if (classField.isInnerClassReference()) {
            return singleton(SyntheticState.SYNTHETIC);
        } else {
            return emptySet();
        }
    }

    private void recordAlias(String className, Class<?> aliasClass) throws IOException {
        removeAlias(className);
        classAliases.put(className, aliasClass);
    }

    private void removeAlias(String className) {
        Class<?> oldAlias = classAliases.remove(className);
        if (oldAlias != null) {
            deleteClassFile(oldAlias.getName());
        }
    }

    private String appendOuterClassNames(String className) {
        Deque<String> parts = new ArrayDeque<>();
        parts.addLast(className);

        String enclosing = enclosingClasses.get(className);
        while (enclosing != null) {
            parts.addFirst(enclosing);
            enclosing = enclosingClasses.get(enclosing);
        }
        return join(parts, "$");
    }

    private String generateClassAlias(String name) {
        return name + GENERATED_BY_TRAFFIC_PARROT + randomId();
    }

    private String randomId() {
        return UUID.randomUUID().toString().replace('-', 'x');
    }

    private boolean hasPartiallyLoadedFields(classdesc classdesc) {
        return any(classFields(classdesc), or(notLoadedObject(), notLoadedArray()));
    }

    private List<String> proxyInterfaces(classdesc classdesc) {
        if (classdesc.interfaces == null) {
            return Collections.emptyList();
        } else {
            return asList(classdesc.interfaces);
        }
    }

    private List<field> classFields(classdesc classdesc) {
        if (classdesc.fields == null) {
            return Collections.emptyList();
        } else {
            return asList(classdesc.fields);
        }
    }

    private Predicate<? super field> notLoadedArray() {
        return new Predicate<field>() {
            @Override
            public boolean apply(field field) {
                return field.type == fieldtype.ARRAY && classAliases.containsKey(className(stripArrayPrefix(field.classname.value)));
            }
        };
    }

    private Predicate<field> notLoadedObject() {
        return new Predicate<field>() {
            @Override
            public boolean apply(field field) {
                return field.type == OBJECT && classAliases.containsKey(className(field.classname.value));
            }
        };
    }

    private Class<?> superClass(classdesc classdesc) throws ClassNotFoundException, IOException {
        classdesc superclass = classdesc.superclass;
        if (superclass != null) {
            return loadClass(superclass, false);
        } else {
            return Object.class;
        }
    }

    private TypeDescription typeOf(String classBeingGenerated, field field) throws ClassNotFoundException, IOException {
        switch (field.type) {
            case BYTE:
                return new ForLoadedType(byte.class);
            case CHAR:
                return new ForLoadedType(char.class);
            case DOUBLE:
                return new ForLoadedType(double.class);
            case FLOAT:
                return new ForLoadedType(float.class);
            case INTEGER:
                return new ForLoadedType(int.class);
            case LONG:
                return new ForLoadedType(long.class);
            case SHORT:
                return new ForLoadedType(short.class);
            case BOOLEAN:
                return new ForLoadedType(boolean.class);
            case ARRAY:
                return arrayType(classBeingGenerated, field);
            case OBJECT:
                return classType(classBeingGenerated, field.classname.value);
            default:
                throw new ClassNotFoundException("Unsupported field type: " + field.type.toString());
        }
    }

    private TypeDescription classType(String classBeingGenerated, String fieldDescriptor) {
        String className = className(fieldDescriptor);
        Class<?> alias = classAliases.get(className);
        if (alias != null) {
            return new ForLoadedType(alias);
        }
        if (classBeingGenerated.equals(className)) {
            return TargetType.DESCRIPTION;
        }
        Optional<? extends Class<?>> alreadyLoadedClass = alreadyLoadedClass(className);
        if (alreadyLoadedClass.isPresent()) {
            return new ForLoadedType(alreadyLoadedClass.get());
        } else {
            return InstrumentedType.Default.of(className, null, 0);
        }
    }

    private String className(String fieldDescriptor) {
        return fieldDescriptor.replaceFirst("L", "").replace('/', '.').replaceFirst(";$", "");
    }

    private String arrayClassName(String className) {
        return "[L" + className + ";";
    }

    private String fieldDescriptor(String className) {
        return "L" + className.replace('.', '/') + ";";
    }

    private String arrayFieldDescriptor(String className) {
        return "[L" + className.replace('.', '/') + ";";
    }

    private TypeDescription arrayType(String classBeingGenerated, field field) throws ClassNotFoundException, IOException {
        String arrayType = field.classname.value;
        switch (arrayType) {
            case "[B":
                return new ForLoadedType(byte[].class);
            case "[C":
                return new ForLoadedType(char[].class);
            case "[D":
                return new ForLoadedType(double[].class);
            case "[F":
                return new ForLoadedType(float[].class);
            case "[I":
                return new ForLoadedType(int[].class);
            case "[J":
                return new ForLoadedType(long[].class);
            case "[S":
                return new ForLoadedType(short[].class);
            case "[Z":
                return new ForLoadedType(boolean[].class);
            default:
                if (arrayType.startsWith("[L")) {
                    String fieldDescriptor = stripArrayPrefix(arrayType);
                    TypeDescription componentDescription = classType(classBeingGenerated, fieldDescriptor);
                    return TypeDescription.ArrayProjection.of(componentDescription);
                } else {
                    throw new ClassNotFoundException("Unsupported array type: " + arrayType);
                }
        }
    }

    private String stripArrayPrefix(String arrayType) {
        checkState(arrayType.startsWith("["));
        return arrayType.substring(1);
    }

    private Optional<? extends Class<?>> alreadyLoadedClass(String name) {
        try {
            return Optional.of(Class.forName(name));
        } catch (ClassNotFoundException e) {
            return Optional.absent();
        }
    }

    private static Object bytesToJavaObject(byte[] bytes) throws IOException, ClassNotFoundException {
        try (ObjectInputStream objectInputStream = new ObjectInputStream(new ByteArrayInputStream(bytes))) {
            return objectInputStream.readObject();
        }
    }
}
